#lang racket/base

(require
 redex/reduction-semantics
 racket/match
 "1-lc-syntax.rkt"
 "2-anf-and-monadic-form.rkt")

(provide (all-defined-out))


;; ------------------------------------------------------------------------
;; Helpers

(define-syntax do
  (syntax-rules ()
    [(do () e) e]
    [(do ([x1 e1] [xs es] ...) e)
     (e1 (lambda (x1)
           (do ([xs es] ...) e)))]))

;; ------------------------------------------------------------------------
;; ANF Translation

;; The not-optimizing ANF translation (only normalizes A reductions, not left identity)
(define (anf e [gensym gensym])
  ;; es -> (List V -> M) -> M
  (define (anf-vs es k)
    (if (null? es)
        (k '())
        ((anf-v (car es))
         (lambda (v)
           (anf-vs (cdr es)
                   (lambda (vs)
                     (k (cons v vs))))))))

  ;; e -> M
  (define (anf-m e)
    ((anf-n e) (lambda (x) x)))

  ;; e -> (V -> M) -> M
  (define ((anf-v e) k)
    ((anf-n e)
     (lambda (n)
       ;; Version to also normalize left identity
       #;(if (value? n)
           (k n)
           (let ([x (gensym 'x)])
             `(let ([,x ,n])
                ,(k x))))
       (let ([x (gensym 'x)])
         `(let ([,x ,n])
            ,(k x))))))

  ;; e -> (N -> M) -> M
  (define ((anf-n e) k)
    (match e
      [(? value?)
       (k e)]
      [`(lambda (,x) ,e)
       (k `(lambda (,x) ,(anf-m e)))]
      [`(apply ,e1 ,e2)
       (define x (gensym 'x))
       (do ([v1 (anf-v e1)]
            [v2 (anf-v e2)])
           `(let ([,x (apply ,v1 ,v2)])
              ,(k x)))]
      [`(,op ,es ...)
       #:when (op? op)
       (anf-vs es (lambda (vs)
                    (k `(,op ,@vs))))]
      ;; Duplication rule
      #;[`(ifz ,e ,e1 ,e2)
       (do ([v (anf-v e)])
           `(ifz ,v
                 ,((anf-n e1) k)
                 ,((anf-n e2) k)))]
      ;; Join point rule
      [`(ifz ,e ,e1 ,e2)
       (define x (gensym 'x))
       (define j (gensym 'j))
       (define k^ (lambda (y) `(apply ,j ,y)))
       (do ([v (anf-v e)])
           `(let ([,j (lambda (,x) ,(k x))])
              (ifz ,v
                   ,((anf-v e1) k^)
                   ,((anf-v e2) k^))))]
      [`(let ([,x ,e1]) ,e2)
       (do ([n (anf-n e1)])
           `(let ([,x ,n])
              ,((anf-n e2) k)))]

      ;; region rules
      [`(letregion ,r ,e)
       `(letregion ,r
                   ,((anf-n e) k))]
      [`(@ ,r ,v)
       #:when (value? v)
       (k `(@ ,r ,v))]
      [`(@ ,r (lambda (,x) ,e))
       (k `(@ ,r (lambda (,x) ,(anf-m e))))]
      [`(@ ,r (,op ,es ...))
       #:when (op? op)
       (anf-vs es (lambda (vs)
                    (k `(@ ,r (,op ,@vs)))))]))
  (anf-m e))


;; ------------------------------------------------------------------------
;; Monadic translation

(define (monadic e [gensym gensym])
  ;; Smart constructor for optimizing the left identity equation
  (define (make-let rhs k)
    (match rhs
      [(? value?)
       (k rhs)]
      [_
       (define x (gensym 'x))
       `(let ([,x ,rhs])
          ,(k x))]))

  (define (make-let* binds e)
    (for/fold ([e e])
              ([bind (reverse binds)])
      `(let (,bind) ,e)))

  ;; internal loop, to keep reference to gensym in closure
  (let monadic ([e e])
    (match e
      [(? value?)
       e]
      [`(lambda (,x) ,e)
       `(lambda (,x) ,(monadic e))]
      [`(apply ,e1 ,e2)
       (define x1 (gensym 'x))
       (define x2 (gensym 'x))
       (make-let*
        `([,x1 ,(monadic e1)]
          [,x2 ,(monadic e2)])
        `(apply ,x1 ,x2))]
      [`(,op ,es ...)
       #:when (op? op)
       (define xs (map (lambda _ (gensym 'x)) es))
       (make-let*
        (for/list ([e es]
                   [x xs])
          `[,x ,(monadic e)])
        `(,op ,@xs))]
      [`(ifz ,e ,e1 ,e2)
       (define x (gensym 'x))
       `(let ([,x ,(monadic e)])
          (ifz ,x
               ,(monadic e1)
               ,(monadic e2)))]
      [`(let ([,x ,e1]) ,e2)
       `(let ([,x ,(monadic e1)])
          ,(monadic e2))]

      ;; The region rules
      [`(letregion ,r ,e)
       `(letregion ,r ,(monadic e))]
      [`(@ ,r ,v)
       #:when (value? v)
       `(@ ,r ,v)]
      [`(@ ,r (lambda (,x) ,e))
       `(@ ,r (lambda (,x) ,(monadic e)))]
      [`(@ ,r (,op ,es ...))
       #:when (op? op)
       (define xs (map (lambda _ (gensym 'x)) es))
       (make-let*
        (for/list ([e es]
                   [x xs])
          `[,x ,(monadic e)])
        `(@ ,r (,op ,@xs)))]
      [`(@ ,r ,e)
       (define x (gensym 'x))
       `(let ([,x ,(monadic e)])
          (@ ,r ,x))])))

(module+ test
  ;; The test produces something in monadic form, but not the same B-normal form
  ;; as the B-normalization.
  ;; They normalize case-of-case in different orders.
  #|
  (let ((x
         (let ((x_2
                (let ((x_1 (ifz 0 0 1)))
                  (ifz x_1 0 1))))
           (ifz x_2 0 1))))
    LARGE)

  However, the two terms code-gen and AB-normalize to the same thing; see later tests
  |#
  (test-no-match monadicL
   (let ([x (let ([x_1 (ifz 0 0 1)])
              (let ([x_2 (ifz x_1 0 1)])
                (ifz x_2
                     0
                     1)))])
     LARGE)
   (monadic nested-branches)))

;;------------------------------------------------------------------------
;; ANF Code generation

(define (anf-code-gen M)
  (define (acg-m m)
    (match m
      [`(let ([,x ,n]) ,m)
       `(begin
          (set! ,x ,(acg-n n))
          ,(acg-m m))]
      [`(ifz ,v ,m1 ,m2)
       `(ifz ,(acg-v v)
             ,(acg-m m1)
             ,(acg-m m2))]

      [`(letregion ,r ,e)
       ;; Okay, I'm cheating. Doing this properly requires more reasoning in CPS.
       ;; But look I've made my point. Stop using ANF.
       (define x (gensym 'x))
       `(begin
          (ralloc ,r)
          (set! ,x ,(acg-m e))
          (rfree ,r)
          ,x)]

      [(? value?)
       (acg-v m)]
      [n (acg-n n)]))

  (define (acg-v v)
    (match v
      [`(lambda (,x) ,m)
       `(lambda (,x) ,(acg-m m))]
      [v v]))

  (define (acg-n n)
    (match n
      [`(@ ,r ,n)
       `(alloc ,r ,(acg-n n))]
      [`(apply ,v1 ,v2)
       `(call ,(acg-v v1) ,(acg-v v2))]
      [`(,op ,vs ...)
       #:when (op? op)
       `(,op ,@(map acg-v vs))]
      [v (acg-v v)]))

  (acg-m M))

(define (monadic-code-gen C)
  (let code-gen ([C C])
    (match C
      [`(let ((,x ,C1)) ,C2)
       `(begin
          (set! ,x ,(code-gen C1))
          ,(code-gen C2))]
      [`(ifz ,U ,C1 ,C2)
       `(ifz ,(code-gen U)
             ,(code-gen C1)
             ,(code-gen C2))]
      [`(,op ,Us ...)
       #:when (op? op)
       `(,op ,@(map code-gen Us))]
      [`(apply ,U1 ,U2)
       `(call ,(code-gen U1) ,(code-gen U2))]
      [`(lambda (,x) ,C)
       `(lambda (,x) ,(code-gen C))]

      [`(letregion ,r ,C)
       (define x (gensym 'x))
       `(begin
          (ralloc ,r)
          (set! ,x ,(code-gen C))
          (rfree ,r)
          ,x)]
      [`(@ ,r ,U)
       `(alloc ,r ,(code-gen U))]
      [U
       U])))


;;------------------------------------------------------------------------
;; AB normalizer
(define (abnf p)
  (define (ab-normalize-t t)
    (match t
      [`(begin ,s ... ,t)
       `(begin ,@(map ab-normalize-s s) ,(ab-normalize-t t))]
      [`(ifz ,v ,t1 ,t2)
       `(ifz ,(ab-normalize-v v)
             ,(ab-normalize-t t1)
             ,(ab-normalize-t t2))]
      [`(,op ,v ...)
       #:when (op? op)
       t]
      [`(call ,v1 ,v2)
       t]
      [v (ab-normalize-v v)]))

  (define (ab-normalize-s is)
    (match is
      [`(begin ,s ...)
       `(begin ,@(map ab-normalize-s s))]
      [`(ifz ,v ,s1 ,s2)
       `(ifz ,(ab-normalize-v v)
             ,(ab-normalize-s s1)
             ,(ab-normalize-s s2))]
      [`(set! ,x (call ,v1 ,v2))
       `(set! ,x (call ,(ab-normalize-v v1) ,(ab-normalize-v v2)))]
      [`(set! ,x (,op ,v ...))
       #:when (op? op)
       is]
      [`(set! ,x (begin ,s ... ,t))
       `(begin ,@(map ab-normalize-s s)
               ,(ab-normalize-s `(set! ,x ,t)))]
      [`(set! ,x (ifz ,v ,t1 ,t2))
       `(ifz ,(ab-normalize-v v)
             ,(ab-normalize-s `(set! ,x ,t1))
             ,(ab-normalize-s `(set! ,x ,t2)))]
      [`(set! ,x ,v)
       #:when (value? v)
       `(set! ,x ,v)]
      [`(set! ,y (lambda (,x) ,t))
       `(set! ,y (lambda (,x) ,(ab-normalize-t t)))]
      ;; region rules
      [`(ralloc ,r)
       is]
      [`(rfree ,r)
       is]
      [`(set! ,y (alloc ,r (lambda (,x) ,t)))
       `(set! ,y (alloc ,r (lambda (,x) ,(ab-normalize-t t))))]
      [`(set! ,x (alloc ,r ,v))
       ;; other
       is]))

  (define (ab-normalize-v v)
    (match v
      [`(lambda (,x) ,t)
       `(lambda (,x) ,(ab-normalize-t t))]
      [_ v]))

  (ab-normalize-t p))

;;------------------------------------------------------------------------
;; Compilers

(define (anf-compile x [gensym gensym])
  (anf-code-gen (anf x gensym)))

(define (abnormal-compile x [gensym gensym])
  (abnf (monadic-code-gen (monadic x gensym))))

(require (only-in racket/syntax format-symbol))
;; A predictable var generator, but may not generate fresh names
(define (inc-var [init 1])
  (let ([count (box init)])
    (lambda (x)
      (define ccount (unbox count))
      (begin0 (format-symbol "~a~a" x ccount)
        (set-box! count (add1 ccount))))))

(module+ test
  (require rackunit redex/reduction-semantics)
  (require "4-ab-normal-form.rkt")

  (test-match
   abnfL
   t
   (abnormal-compile running-example (inc-var)))

  (test-match
   abnfL
   t
   (anf-compile running-example (inc-var)))

  ;; both compilers produce equal AB normal forms for programs without bind/if commuting
  ;; (up to flattening and alpha-equivalence)
  ;; They're not quite syntactically equal
  #;(check equal?
   (flatten-begin (anf-compile running-example (inc-var)))
   (flatten-begin (abnormal-compile running-example (inc-var))))

  (test-match imonadicL
   (begin
     (begin (ifz 0 (set! x_1 0) (set! x_1 1))
            (begin (ifz x_1 (set! x_2 0) (set! x_2 1))
                   (ifz x_2 (set! x 0) (set! x 1))))
     LARGE)
   (ab-normalize
    (monadic-code-gen
     (b-normalize
      nested-branches))))

  ;; monadic translation reaches same normal form as B-normalization then code gen
  (test-match
   imonadicL
   (begin
     (begin
       (begin
         (begin
           (set! x_3 0)
           (ifz x_3 (set! x_1 0) (set! x_1 1)))
         (ifz x_1 (set! x_2 0) (set! x_2 1)))
       (ifz x_2 (set! x 0) (set! x 1)))
     LARGE)
   (ab-normalize
    (monadic-code-gen
     (monadic
      nested-branches))))

  ;; stress tests
  (test-match
   monadicL
   C
   (monadic fact-example))

  (test-match
   imonadicL
   t
   (monadic-code-gen (monadic fact-example)))

  (test-match
   abnfL
   t
   (abnf (monadic-code-gen (monadic fact-example))))

  (test-match
   ANFL
   M
   (anf fact-example))

  (test-match
   abnfL
   t
   (anf-code-gen (anf fact-example))))
