#lang scribble/manual
@(require
  scribble/examples
  (for-label
   "main.rkt"
   racket/base
   redex/reduction-semantics
   redex/gui
   racket/list
   racket/contract))

@title{Artifact: A low-level look at A-normal Form}
@author{Anon Y. Mous}

@(define e (make-base-eval '(require racket/list)))

@defmodule["main.rkt" #:packages ()]{
This artifact requires Racket, and has been tested in Racket 8.7 and above.
An installer can can be downloaded for most major systems at @url{https://download.racket-lang.org}.

It relies on the Redex package, which is included in the standard Racket install, but can be installed manually using @tt{raco pkg install redex}.

The main entry point for this artifact, which provides all languages, examples,reduction relations, and compilers defined in the paper.
This is largely implemented in @racketmodname[redex], but this documentation tries to provide sufficient usage instructions that knowledge of Redex is not required.

You can use the artifact by importing "main.rkt" in the REPL or any other module with @racket[(require "main.rkt")], or running @tt{racket -i -t main.rkt}, or opening "main.rkt" in DrRacket.
It may take a minute as importing this module will compile and run the test suite.

There are minor differences between the paper and the implementation:
@itemize[
@item{The syntax for function application uses the keyword @tt{apply} rather
than juxtaposition.}
@item{@tt{let} requires a second pair of parenthesis, as in @racket[(let ([x 5]) x)].}
]

@examples[#:eval e
(require "main.rkt")
a->
anf
abnormal-compile
anf-compile
(anf '(apply (lambda (x) x) 5))
(anf '(apply (lambda (x) x) 5) (inc-var))
]
}

@section{λ-calculus, intro, examples, etc.}
@defthing[#:kind "language" lambdaL language?]{
The syntax of the λ-calculus, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(redex-match?
 lambdaL
 e
 (term (apply (lambda (x) 5) 6)))

(redex-match?
 lambdaL
 ι
 '(apply (lambda (x) 5) 6))

(redex-match?
 lambdaL
 ι
 5)

(redex-match?
 lambdaL
 x
 'x1)
]

}

@defthing[#:kind "language" eval-lambdaL language?]{
The syntax of the λ-calculus extended with evaluation contexts and values, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(redex-match?
 eval-lambdaL
 (in-hole E v)
 '(apply (lambda (x) 5) 6))

(redex-match
 eval-lambdaL
 (in-hole E v)
 '(apply (lambda (x) 5) 6))

(redex-match?
 eval-lambdaL
 E
 '(apply hole 6))
]
}

@defthing[#:kind "term" intro-example (redex-match? lambdaL e)]{
The λ-calculus example from the intro.
}

@defthing[#:kind "term" running-example (redex-match? lambdaL e)]{
The λ-calculus running example.
}

@defthing[#:kind "term" nested-branches-example (redex-match? lambdaL e)]{
The λ-calculus example with nested case-of-case pattern.

@examples[#:eval e
nested-branches-example
]
}

@defthing[#:kind "term" fact-example (redex-match? lambdaL e)]{
A definition of the factorial function in the λ-calculus.
}

@defthing[#:kind "language" lambda-ckL language?]{
The syntax of the λ CK machine as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? lambda-ckL K 'mt) #t)
(eval:check (redex-match? lambda-ckL K (term ((+ 4 hole) :: mt))) #t)
]

@margin-note{In Redex, @racket[term] behaves almost the same as @racket[quote],
but there are exceptions. For example, you must use @racket[term] to write a
term with a @racket[hole].}
}

@defproc[(init-ck [term (redex-match? lambdaL e)])
         (redex-match? lambda-ckL (e K))]{
Produces an initial machine configuration for the λ CK machine using @racket[term] for the initial code.
@examples[#:eval e
(init-ck intro-example)
]
}

@defthing[#:kind "reduction-relation" lambda-ck-> reduction-relation?]{
The λ CK abstract machine, as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

@examples[#:eval e
(apply-reduction-relation* lambda-ck-> (init-ck intro-example))
(car (car (apply-reduction-relation* lambda-ck-> (init-ck fact-example))))
]
}

@section{A-normal form and Monadic form}
@defthing[#:kind "reduction-relation" a-> reduction-relation?]{
The single-step A-reductions for the @racket[lambdaL], as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

This is non-determinisic due to fresh name generation.
The empty set of answers is returned when the program cannot single-step at the top-level.

@examples[#:eval e
(apply-reduction-relation a-> intro-example)
(car (apply-reduction-relation a-> intro-example))
]
}

@defthing[#:kind "reduction-relation" a->* reduction-relation?]{
The compatible closure of the @racket[a->], as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.

@examples[#:eval e
(apply-reduction-relation a->* intro-example)
(apply-reduction-relation* a->* #:cache-all? #t nested-branches-example)
]
}

@defproc[(a-normalize [term (redex-match? lambdaL e)])
         (redex-match? ANFL m)]{
A-normalizes @racket[term] by running @racket[a->*] to a normal form.
@examples[#:eval e
(a-normalize intro-example)
]
}

@defthing[#:kind "language" ANFL language?]{
The syntax of A-normal form, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(redex-match? ANFL M '(let ([x (apply f 1)]) (+ x 2)))
(redex-match? ANFL M '(+ (apply f 1) 2))
]
}


@defthing[#:kind "language" monadic-eval-lambdaL language?]{
The syntax of the λ-calculus extendedwith non-strict monadic evaluation contexts and values, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(redex-match? monadic-eval-lambdaL En (term (let ([x hole]) 5)))
(redex-match? eval-lambdaL E (term (let ([x hole]) 5)))

(redex-match? monadic-eval-lambdaL En (term (ifz hole 5 6)))
(redex-match? eval-lambdaL E (term (ifz hole 5 6)))
]
}

@defthing[#:kind "reduction-relation" b-> reduction-relation?]{
The single-step B-reductions for the @racket[lambdaL], as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.
The empty set of answers is returned when the program cannot single-step at the top-level.

@examples[#:eval e
(apply-reduction-relation b-> intro-example)
(apply-reduction-relation b-> nested-branches-example)
]
}

@defthing[#:kind "reduction-relation" b->* reduction-relation?]{
The compatible closure of the @racket[b->], as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.

@examples[#:eval e
(car (apply-reduction-relation b->* nested-branches-example))
(car (apply-reduction-relation* b->* nested-branches-example))
]
}

@defproc[(b-normalize [term (redex-match? lambdaL e)])
         (redex-match? monadicL C)]{
B-normalizes @racket[term] by running @racket[b->*] to a normal form.
@examples[#:eval e
(b-normalize nested-branches-example)
]
}

@defthing[#:kind "language" monadicL language?]{
The syntax of B-normal form, @emph{i.e.}, monadic form, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(redex-match? monadicL C '(let ([x (apply f 1)]) (+ x 2)))
(redex-match? monadicL C '(+ (apply f 1) 2))
(redex-match? monadicL C '(let ((y (let ([x (apply f 1)]) (+ x 2)))) y))
]
}

@section{ANF and Monadic Machines}

@defthing[#:kind "reduction-relation" anf-ck-> reduction-relation?]{
The ANF CK abstract machine, as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

@examples[#:eval e
(apply-reduction-relation* lambda-ck-> (init-ck running-example))
(apply-reduction-relation* anf-ck-> (init-ck (a-normalize running-example)))
]
}

@defproc[(ck-max-stack [states (listof (redex-match? lambda-ckL (e K)))])
         natural-number?]{
Returns the maximum stack length found in a CK machine trace (either @racket[lambda-ck->] or @racket[anf-ck->]).

@examples[#:eval e
(ck-max-stack (apply-reduction-relation* #:all? #t anf-ck-> (init-ck (a-normalize running-example))))
(ck-max-stack (apply-reduction-relation* #:all? #t lambda-ck-> (init-ck running-example)))
]
}

@defthing[#:kind "language" regionL language?]{
The λ-region syntax, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? regionL e '(\@ r0 5)) #t)
(eval:check (redex-match? regionL e '(letregion r1 (\@ r0 5))) #t)
]
}

@defthing[#:kind "language" regionCSKL language?]{
The λ-region CSK abstract machine syntax, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? regionCSKL K (term ((free r1) :: mt))) #t)
(eval:check (redex-match? regionCSKL S '()) #t)
(eval:check (redex-match? regionCSKL S '(() (r0 ()))) #t)
(eval:check (redex-match? regionCSKL S '(() (r0 (() (o1 5))))) #t)
(eval:check (redex-match? regionCSKL a '(r0 o1)) #t)
]

@margin-note{In Redex, @racket[term] behaves almost the same as @racket[quote],
but there are exceptions. For example, you must use @racket[term] to write a
term with a @racket[hole].}
}

@defthing[#:kind "reduction-relation" region-csk-> reduction-relation?]{
The λ-region CSK abstract machine, as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

@examples[#:eval e
(csk-read-val (apply-reduction-relation* region-csk-> (init-csk reg1-term)))
]
}

@defthing[reg1-term (redex-match? regionL e)]{
An example term for λ-regions.
}

@defproc[(init-csk [term (redex-match? regionL e)])
         (redex-match? regionCSKL (e S K))]{
Produces an initial machine configuration for the λ-region CSK machine using @racket[term] for the initial code, and an empty initial store and kontinuation.

@examples[#:eval e
(init-csk reg1-term)
]
}

@defproc[(csk-read-val [states (listof (redex-match? regionCSKL (e S K)))])
         (redex-match? regionL v)]{
Produces the final value of the CSK machine from its final configuration.
Expects that @racket[states] is a singleton set of terminal machine configurations, @emph{i.e.}, that the machine normalized to a single final state, whose code is a valid address and whose kontinuation is empty.

@examples[#:eval e
(csk-read-val '(((r0 o1) (() (r0 (() (o1 5)))) mt)))
]
}

@defproc[(csk-max-regions [trace (listof (redex-match? regionCSKL (e S K)))])
         natural-number?]{
Returns the maximum number of regions allocated in the @racket[region-csk->] trace.

@examples[#:eval e
(csk-max-regions (apply-reduction-relation* region-csk-> (init-csk reg1-term)))
(csk-max-regions (apply-reduction-relation* region-csk-> (init-csk anf-reg1-term)))
(csk-max-regions (apply-reduction-relation* region-csk-> (init-csk bnf-reg1-term)))
]
}

@defproc[(csk-max-memory [trace (listof (redex-match? regionCSKL (e S K)))])
         natural-number?]{
Returns the maximum number of live addresses in the @racket[region-csk->] trace.

@examples[#:eval e
(csk-max-memory (apply-reduction-relation* region-csk-> (init-csk reg1-term)))
(csk-max-memory (apply-reduction-relation* region-csk-> (init-csk anf-reg1-term)))
(csk-max-memory (apply-reduction-relation* region-csk-> (init-csk bnf-reg1-term)))
]
}

@defthing[#:kind "language" aregionL language?]{
The λ-regions syntax extended with evaluation contexts for A-reduction, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? aregionL E (term (\@ r hole))) #t)
(eval:check (redex-match? aregionL E (term (let ([x hole]) e))) #t)
]

@margin-note{In Redex, @racket[term] behaves almost the same as @racket[quote],
but there are exceptions. For example, you must use @racket[term] to write a
term with a @racket[hole].}
}

@defthing[#:kind "reduction-relation" ra-> reduction-relation?]{
The single-step A-reductions for λ-regions (@racket[aregionL]), as a Redex @racket[reduction-relation].

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.
The empty set of answers is returned when the program cannot single-step at the top-level.

@examples[#:eval e
(apply-reduction-relation ra-> reg1-term)
]
}

@defthing[#:kind "reduction-relation" ra->* reduction-relation?]{
The compatible closure of @racket[ra->], as a Redex @racket[reduction-relation].

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.
This can be rather slow due to non-determinism, so you probably want to enable @racket[#:cache-all?].

@examples[#:eval e
(take (apply-reduction-relation ra->* reg1-term) 3)
(car (apply-reduction-relation* ra->* #:cache-all? #t reg1-term))
]
}

@defthing[anf-reg1-term (redex-match? regionL e)]{
The A-normal form of @racket[reg1-term]
}

@defthing[#:kind "language" bregionL language?]{
The λ-regions syntax extended with non-strict monadic evaluation contexts for B-reduction, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? bregionL En (term (\@ r hole))) #t)
(eval:check (redex-match? bregionL En (term (let ([x hole]) e))) #f)
]

@margin-note{In Redex, @racket[term] behaves almost the same as @racket[quote],
but there are exceptions. For example, you must use @racket[term] to write a
term with a @racket[hole].}
}

@defthing[#:kind "reduction-relation" rb-> reduction-relation?]{
The single-step B-reductions for the λ-regions (@racket[aregionL]), as a Redex @racket[reduction-relation].

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.
The empty set of answers is returned when the program cannot single-step at the top-level.

@examples[#:eval e
(apply-reduction-relation rb-> reg1-term)
]
}

@defthing[#:kind "reduction-relation" rb->* reduction-relation?]{
The compatible closure of @racket[rb->], as a Redex @racket[reduction-relation].

This is non-determinisic due to fresh name generation and (confluent) choices of evaluation contexts.
This can be rather slow due to non-determinism, so you probably want to enable @racket[#:cache-all?].

@examples[#:eval e
(take (apply-reduction-relation rb->* reg1-term) 3)
(car (apply-reduction-relation* rb->* #:cache-all? #t reg1-term))
]
}

@defthing[bnf-reg1-term (redex-match? regionL e)]{
The B-normal form of @racket[reg1-term]
}

@section{Imperative A-normalization, Machines, and AB-normalization}

@defthing[#:kind "language" imonadicL language?]{
The imperative monadic syntax, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? imonadicL t '(begin (set! x (begin (set! y 5) y)) x)) #t)
]
}

@defthing[#:kind "language" abnfL language?]{
The imperative ANF syntax, @emph{i.e.}, AB-normal form, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? imonadicL t '(begin (set! x (begin (set! y 5) y)) x)) #t)
(eval:check (redex-match? abnfL t '(begin (set! x (begin (set! y 5) y)) x)) #f)
(eval:check (redex-match? abnfL t '(begin (set! y 5) (set! x y) x)) #t)
]
}

@defthing[#:kind "reduction-relation" ab-> reduction-relation?]{
The single-step AB-reductions for the @racket[imonadicL] @emph{statements}, as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

The empty set of answers is returned when the program cannot single-step at the top-level.

@examples[#:eval e
(apply-reduction-relation ab-> '(set! x (begin (set! y 5) y)))
]
}

@defthing[#:kind "reduction-relation" ab->* reduction-relation?]{
The compatible closure of @racket[ab->], as a Redex @racket[reduction-relation].
This lifts @racket[ab->] to work over tails, as well.
Try using with @racket[traces] to navigate the trace graphically.

@examples[#:eval e
(apply-reduction-relation ab->* '(begin (set! x (begin (set! y 5) y)) x))
(apply-reduction-relation ab->* (monadic-code-gen (b-normalize nested-branches-example)))
]
}

@defproc[(ab-normalize [term (redex-match? imonadicL t)])
         (redex-match? abnfL t)]{
AB-normalizes @racket[term] by running @racket[ab->*] to a normal form.
@examples[#:eval e
(ab-normalize (monadic-code-gen (b-normalize nested-branches-example)))
]
}

@defthing[#:kind "language" imonadic-cekL language?]{
The imperative CEK abstract machine syntax, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? imonadic-cekL K (term ((begin (set! x hole) x) :: mt))) #t)
(eval:check (redex-match? imonadic-cekL Σ (term (() (x 5)))) #t)
]

@margin-note{In Redex, @racket[term] behaves almost the same as @racket[quote],
but there are exceptions. For example, you must use @racket[term] to write a
term with a @racket[hole].}
}

@defthing[#:kind "reduction-relation" cek-> reduction-relation?]{
The imperative CEK abstract machine, as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

@examples[#:eval e
(apply-reduction-relation* cek-> (init-cek '(begin (set! x (begin (set! y 5) y)) x)))
(car (car (apply-reduction-relation* cek-> (init-cek '(begin (set! x (begin (set! y 5) y)) x)))))
]
}

@defproc[(init-cek [term (redex-match? imonadicL e)])
         (redex-match? imonadic-cekL (e Σ K))]{
Produces an initial machine configuration for the monadic imperative CEK machine using @racket[term] for the initial code.
@examples[#:eval e
(init-cek '(begin (set! x 5) x))
]
}

@defproc[(cek-max-stack [states (listof (redex-match? imonadic-cekL (e Σ K)))])
          natural-number?]{
Returns the maximum stack length found in a @racket[cek->] trace.

@examples[#:eval e
(cek-max-stack
 (apply-reduction-relation* #:all? #t
  cek->
  (init-cek (monadic-code-gen (b-normalize nested-branches-example)))))
(cek-max-stack
 (apply-reduction-relation* #:all? #t
  cek->
  (init-cek (ab-normalize (monadic-code-gen (b-normalize nested-branches-example))))))
]
}

@defthing[#:kind "language" rimonadicL language?]{
The imperative monadic with regions syntax, as a Redex language (see @racket[define-language]).

@examples[#:eval e
(eval:check (redex-match? rimonadicL t '(begin (set! x (begin (set! y 5) y)) x)) #f)
(eval:check (redex-match? rimonadicL t '(begin (set! x (begin (set! y (alloc r0 5)) y)) x)) #t)
]
}

@defthing[#:kind "reduction-relation" cesk-> reduction-relation?]{
The imperative CESK abstract machine, as a Redex @racket[reduction-relation].
Try using with @racket[traces] to navigate the trace graphically.

@examples[#:eval e
(apply-reduction-relation* cesk-> (init-cesk '(begin (set! x (begin (set! y (alloc r0 5)) y)) x)))
(cesk-read-val (apply-reduction-relation* cesk-> (init-cesk '(begin (set! x (begin (set! y (alloc r0 5)) y)) x))))
]
}

@defproc[(init-cesk [term (redex-match? rimonadicL t)])
         (redex-match? rimonadicL (t Σ S K))]{
Produces an initial machine configuration for the monadic imperative with regions CESK machine using @racket[term] for the initial code.
The machine starts with an empty environment, empty kontinuation, and an initial empty region @racket[r0] where the final result must be allocated.

@examples[#:eval e
(init-cesk '(begin (set! x (alloc r0 5)) x))
]
}

@defproc[(cesk-read-val [states (listof (redex-match? rimonadicL (t Σ S K)))])
         (redex-match? rimonadicL v)]{
Produces the final value of the CESK machine from its final configuration.
Expects that @racket[states] is a singleton set of terminal machine configurations, @emph{i.e.}, that the machine normalized to a single final state, whose code is a valid address and whose kontinuation is empty.

@examples[#:eval e
(cesk-read-val (apply-reduction-relation* cesk-> (init-cesk '(begin (set! x (alloc r0 5)) x))))
]
}

@defproc[(cesk-max-stack [states (listof (redex-match? rimonadicL (t Σ S K)))])
         natural-number?]{
Returns the maximum stack length found in a @racket[cesk->] trace.

@examples[#:eval e
(cesk-max-stack (apply-reduction-relation* #:all? #t cesk-> (init-cesk '(begin (set! x (alloc r0 5)) x))))
(define D_b
 (apply-reduction-relation* #:all? #t cesk->
  (init-cesk (monadic-code-gen bnf-reg1-term))))
(define D_a
 (apply-reduction-relation* #:all? #t cesk->
  (init-cesk (anf-code-gen anf-reg1-term))))
(define D_ab
 (apply-reduction-relation* #:all? #t cesk->
  (init-cesk (abnf (monadic-code-gen bnf-reg1-term)))))
(cesk-max-stack D_b)
(cesk-max-stack D_a)
(cesk-max-stack D_ab)
]

@margin-note{A minor bug exists in the ANF code generator for regions which unnecessarily introduces a stack frame. Note that this bug does not affect the @racket[cek-max-stack].}
}

@defproc[(cesk-max-memory [states (listof (redex-match? rimonadicL (t Σ S K)))])
         natural-number?]{
Returns the maximum number of live addresses in the @racket[cesk->] trace.

@examples[#:eval e
(cesk-max-memory (apply-reduction-relation* #:all? #t cesk-> (init-cesk '(begin (set! x (alloc r0 5)) x))))
(cesk-max-memory D_b)
(cesk-max-memory D_a)
(cesk-max-memory D_ab)
]
}

@defproc[(cesk-max-regions [states (listof (redex-match? rimonadicL (t Σ S K)))])
         natural-number?]{
Returns the maximum number of regions allocated in the @racket[cesk->] trace.

@examples[#:eval e
(cesk-max-regions (apply-reduction-relation* #:all? #t cesk-> (init-cesk '(begin (set! x (alloc r0 5)) x))))
(cesk-max-regions D_b)
(cesk-max-regions D_a)
(cesk-max-regions D_ab)
]
}

@section{Compiler}
@defproc[(anf [term (or/c (redex-match? regionL e) (redex-match? lambdaL e))]
              [fresh (-> symbol? symbol?) gensym])
         (or/c (redex-match? regionL e) (redex-match? ANFL M))]{
Produces the A-normal form of a λ-calculus or λ-regions term.
Optionally takes a source of fresh names.

@examples[#:eval e
(anf intro-example)
(anf reg1-term)
(anf intro-example (inc-var))
]
}

@defproc[(monadic [term (or/c (redex-match? regionL e) (redex-match? lambdaL e))]
                  [fresh (-> symbol? symbol?) gensym])
         (or/c (redex-match? regionL e) (redex-match? monadicL C))]{
Produces the monadic form of a λ-calculus or λ-regions term.

@examples[#:eval e
(monadic intro-example)
(monadic reg1-term)
(monadic reg1-term (inc-var))
]
}

@defproc[(anf-code-gen [term (or/c (redex-match? regionL e) (redex-match? ANFL M))])
         (or/c (redex-match? rimonadicL t) (redex-match? abnfL t))]{
Genernate the CESK machine code of the ANF term @racket[term].
The output is AB-normal form if the input was A-normal.

@examples[#:eval e
(flatten-begin (anf-code-gen (anf intro-example)))
(flatten-begin (anf-code-gen (anf reg1-term (inc-var))))
]

@margin-note{A minor bug exists in the ANF code generator for regions which unnecessarily introduces a stack frame. Note that this bug does not affect the @racket[cek-max-stack].}
}
}

@defproc[(monadic-code-gen [term (or/c (redex-match? regionL e) (redex-match? monadicL C))])
         (or/c (redex-match? rimonadicL t) (redex-match? imonadicL t))]{
Genernate the CESK machine code of the monadic term @racket[term].

@examples[#:eval e
(flatten-begin (monadic-code-gen (monadic intro-example)))
(flatten-begin (monadic-code-gen (monadic reg1-term)))
]
}

@defproc[(abnf [term (or/c (redex-match? rimonadicL t) (redex-match? imonadicL t))])
         (or/c (redex-match? rimonadicL t) (redex-match? abnfL t))
]{
AB-normalize the CESK machine code of the monadic term @racket[term].

@examples[#:eval e
(abnf (monadic-code-gen (monadic intro-example)))
(flatten-begin (abnf (monadic-code-gen (monadic reg1-term))))
(flatten-begin (anf-code-gen (anf reg1-term)))
(flatten-begin (abnf (anf-code-gen (anf reg1-term))))
(flatten-begin (anf-code-gen (anf intro-example)))
(flatten-begin (abnf (anf-code-gen (anf intro-example))))
]
}

@defproc[(anf-compile [term (or/c (redex-match? regionL e) (redex-match? lambdaL e))]
                      [fresh (-> symbol? lsymbol?)])
         (or/c (redex-match? rimonadic t) (redex-match? abnfL t))]{
Compile a term with the ANF compiler.
Optionally takes a source of fresh names.

@examples[#:eval e
(flatten-begin (anf-compile nested-branches-example (inc-var)))
]
}

@defproc[(abnormal-compile [term (or/c (redex-match? regionL e) (redex-match? lambdaL e))]
                           [fresh (-> symbol? symbol?) gensym])
         (or/c (redex-match? rimonadic t) (redex-match? abnfL t))]{
Compile a term with the AB-normal compiler.
Optionally takes a source of fresh names.

@examples[#:eval e
(flatten-begin (abnormal-compile nested-branches-example (inc-var)))
]
}

@defproc[(inc-var [init natural? 1])
         (-> symbol? symbol?)]{
A predictable name generator factory.
Optionally takes an starting number to append to a generated name.

May not actually generate fresh names, depending on the context in which it is used.

@examples[#:eval e
(define gsym (inc-var 2))
(gsym 'x)
(gsym 'x)
]
}

@defproc[(flatten-begin [term any/c]) any/c]{
Flatten all nested begins; normalizes the @tt{admin1} and @tt{admin2} transitions.
Not used in any of the compilers; useful for pretty-printing.

@examples[#:eval e
(flatten-begin '(begin (begin (set! x y) (begin (set! x 5))) x))]
}
